/*
 * File whichrec.c
 *
 * Authors Philip Duval (DESY)
 *
 * Copyright (c) Deutsches Elektronen-Synchrotron DESY 1992-2023
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * You may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <stdio.h>
#include <stddef.h>
#include <stdlib.h>
#include <sys/types.h>
#include "tine.h"
#include "archive.h"

extern int gAllocDataBuffers;
int getEntryFromRecno(int rec,int nrecs);
int getNumLines(FILE *fp);
DAQRecord *daqTbl;
int totalNrRecords, nrecords;
int nadbfiles = 0;
int archdebug = 0,archPending = 0;
int recordsCapacity = 0;
int g_isTerminating = 0;
int AppendProblemsLog(char *s) { return 0;}
char dbFileName[256];
int getNumActiveRecords(char *fname);

int getDatabaseInfo(char *dbf)
{
  FILE *fp=NULL;
  char str[256];
  int done = 0,cc = 0;
  int database_col;

  if ((fp=fopen("../bin/database.csv","rt")) == NULL && 
      (fp=fopen("../bin/startup.csv","rt")) == NULL)
  {
    strcpy(dbf,"datam.csv");
    return 0;
  }
  while (fgets(str,255,fp) != NULL )
  {
    if (strlen(str) == 0) continue;
    if (strchr("%;#\n",str[0])) continue;      /* comment line(s) */
    if (!done)
    {
      if ((database_col=findcol(str,"DATABASE")) < 0) ccerr(no_such_line);
      done = 1;
      continue;
    }
    strncpy(dbf,colptr(database_col,str),170);
    break;
  }
err:
  if (fp != NULL) fclose(fp);
  return cc;
}
typedef struct RECFILESIZE
{
  int idx;
  long siz;
  struct RECFILESIZE *nxt;
} RECFILESIZE;
RECFILESIZE *RecSzList = NULL;
void freeRecSzList(void)
{
  RECFILESIZE *r = RecSzList, *rnxt = NULL;
  while (r != NULL)
  {
    rnxt = r->nxt;
    free(r);
    r = rnxt;
  }
  RecSzList = NULL;
}
// large to small ...
void insertIntoList(RECFILESIZE *rsiz)
{
  RECFILESIZE *r, *rprv = NULL;
  if (rsiz->siz == 0) return;
  if (RecSzList == NULL)
  { // no list yet !
    RecSzList = rsiz;
    return;
  }
  for (r=RecSzList; r != NULL; r = r->nxt)
  {
    if (r->siz < rsiz->siz)
    {
      rsiz->nxt = r;
      if (r == RecSzList)
      {
        RecSzList = rsiz;
      }
      else
      {
        rprv->nxt = rsiz;
      }
      return;
    }
    rprv = r;
  }
  // we're at the end !
  if (rprv != NULL) rprv->nxt = rsiz;
}
char *usageText = "usage: dbscanfiles /y=<year(4 characters)> /m=<month (1 to 12)> /d=<depth (default=10)>";
int main(int argc, char *argv[])
{
  time_t now = time(NULL);
  struct tm *lt = localtime(&now);
  int i,n,m,k,base = 10,recno,cc = 0;
  int yy=lt->tm_year,mm=lt->tm_mon+1,mstart=1,mstop=12,depth=10;
  int this_year=yy,this_month=mm;
  char *c, yr[256], fn[256], afd[256], afn[32], scratch[72], cwd[256], key[256], prms[512];
  struct stat sbuf;
  static char adbp[256];
  RECFILESIZE *rsiz;

  sprintf(yr,"%d",yy + 1900);
  for (i = 1; i < argc; i++)
  { /* options given */
    c = argv[i];
    if (*c++ != '/') break;
    switch (tolower(*c))
    {
      case 'y': // target year 
        if ((c=strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(yr,c,64);
          if (strlen(yr) != 4)
          {
            printf("enter full year: e.g. 1999 or 2022\n");
            exit(0);
          }
        }
        continue;
      case 'm': // target month
        if ((c = strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          mm = atoi(scratch);
          if (mm < 0) mm = 1;
          if (mm > 12) mm = 12;
          mstart = mstop = mm;
        }
        continue;
      case 'd': // depth
        if ((c = strchr(c,'=')) != NULL)
        {
          c++;
          strncpy(scratch,c,64);
          depth = atoi(scratch);
        }
        continue;
      case '?':
      case 'h': // help
        printf("%s\n",usageText);
        exit(0);
      default:
        continue;
    }
  }

  yy = atoi(&yr[2]);
  if (yy < 60) yy += 100;
  recno = strtol(c,NULL,base);
  gAllocDataBuffers = FALSE;
  if ((cc=getDatabaseInfo(dbFileName)) != 0) ccerr(cc);
  if ((c=getenv("PWD")) != NULL) strncpy(cwd,c,256); // want to gloss over symbolic links !

  if ((c=strrchr(cwd,FS_DELIMITER)) == NULL) ccerr(file_error);
  *c = 0;
  sprintf(fn,"%.128s/database/%.32s",cwd,dbFileName);
  if (stat(fn, &sbuf) != 0) ccerr(no_such_file);  
  sprintf(adbp,"%.128s/database/",cwd);
  ArchiveDBpath = adbp;
  totalNrRecords = getNumActiveRecords(dbFileName);
  recordsCapacity = totalNrRecords;
  if ((daqTbl=(DAQRecord *)calloc(totalNrRecords,sizeof(DAQRecord))) == NULL) ccerr(out_of_server_memory);
  if ((prpAlias=(PRP_ALIAS *)calloc(recordsCapacity,sizeof(PRP_ALIAS))) == NULL) ccerr(out_of_server_memory);
  if ((subSystems=(SUB_SYS *)calloc(recordsCapacity,sizeof(SUB_SYS))) == NULL) ccerr(out_of_server_memory);

  n = loadDaqDb(dbFileName,0);
  loadDaqParamDb(n,0);

  for (m = mstart; m <= mstop; m++)
  {
    freeRecSzList();
    sprintf(afd,"%.128s/DATA/%.4s/%02d/",cwd,yr,m);
    for (i = 0; i < totalNrRecords; i++)
    {
      sprintf(afn,"ta%02d%02d01.%x",yy,m,daqTbl[i].rec);
      sprintf(fn,"%.200s%.28s",afd,afn);
      if (stat(fn,&sbuf) == 0)
      {
        if ((rsiz=(RECFILESIZE *)calloc(1,sizeof(RECFILESIZE))) == NULL) { printf("Houston, we've got a problem!\n");  exit(1); }
        rsiz->siz = sbuf.st_size;
        rsiz->idx = i;
        insertIntoList(rsiz);
      }
    }
    if (yy >= this_year && m > this_month) break;
    printf("year: %.4s, month: %d, top %d largest files:\n",yr,m,depth);
    printf("%-48s %-40s %-20s %-5s %s\n","Address","Keyword(s)","Record No. (hex)","Index"," Size (bytes)");
    for (n = 0, rsiz = RecSzList; n < depth && rsiz != NULL; n++, rsiz=rsiz->nxt)
    {
      sprintf(key,"/%.32s/%.32s/%.64s[%.64s]",daqTbl[rsiz->idx].ctx,daqTbl[rsiz->idx].srv,daqTbl[rsiz->idx].dev,daqTbl[rsiz->idx].prp);
      prms[0] = 0;
      for (k = 0; k < daqTbl[rsiz->idx].nparam; k++)
      {
        if (strlen(prms) < 256)
        {
          if (k > 0) strcat(prms,", ");
          strncat(prms,daqTbl[rsiz->idx].daqParam[k].param,60);
        }
      }
      sprintf(scratch,"(0x%x)",daqTbl[rsiz->idx].rec);
      printf("%-48.48s %-40.40s %4d   %9s %9d  %12ld\n",
             key,prms,daqTbl[rsiz->idx].rec,scratch,rsiz->idx,rsiz->siz);
    }
  }

err:
  if (cc != 0)
  {
    printf("ERROR: %.32s\n",cc2str(cc));
  }
  return 0;
}
int getEntryFromRecno(int rec,int nrecs)
{
  int i;

  for (i=0; i<nrecs; i++)
    if (daqTbl[i].rec == rec) return i;
  return -1;
}
